from os.path import exists

import pandas as pd
import numpy as np
from sentence_transformers import SentenceTransformer
from OLD.candaitelistgenerator import groupNifDocumentByRefContext, load
import json
import tensorflow as tf
import tensorflow.keras.models
from tensorflow.keras.layers import Input, Dense, Dropout, concatenate, ReLU, Flatten
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.models import Model
from tensorflow.keras.losses import CosineSimilarity, MeanSquaredError
from tensorflow.keras import backend as K
from sklearn.metrics import f1_score
import time

from OLD.constants import dataset_names
config = tf.compat.v1.ConfigProto(gpu_options = tf.compat.v1.GPUOptions(per_process_gpu_memory_fraction=0.9)
# device_count = {'GPU': 1}
)
config.gpu_options.allow_growth = True
session = tf.compat.v1.Session(config=config)
tf.compat.v1.keras.backend.set_session(session)

tensorflow.random.set_seed(2022)
sentence_embedding_model = SentenceTransformer("all-mpnet-base-v2")


def prepare_embeddings():
    embeddings_df = pd.read_csv("../data/embedding_subset.tsv", sep="\t")
    embeddings_dict = {}
    for index, entry in embeddings_df.iterrows():
        embeddings_dict[entry.get('entity')] = entry.get('embeddings')
    return embeddings_df, embeddings_dict


def prepare_sentence_embeddings(sentences):
    sentence_embeddings = {}
    embeddings = sentence_embedding_model.encode(sentences)
    for sentence, embedding in zip(sentences, embeddings):
        sentence_embeddings[sentence] = embedding
    return sentence_embeddings


def preparedocumentcontextmap(entity_embeddings, data):
    documentcontextmap = {}
    for i in data:
        docid = i
        documentvectors = []
        for mention in data[i]:
            embed_vectors = []
            for candidate in mention['candidates']:
                if candidate in entity_embeddings:
                    embed_vectors.append(np.array(entity_embeddings[candidate].split(","), dtype='float'))
                else:
                    embed_vectors.append(np.full(100, -1))
            documentvectors.append(embed_vectors)
        documentcontextmap[i] = documentvectors
    return documentcontextmap


def prepare_training_data(entity_embeddings, ds_name):
    sentences = []
    x_train = []
    y_train = []
    doc_embeddings = []
    comtextmatrices = []
    document_dictionary = groupNifDocumentByRefContext(load("./data/nifData/" + ds_name))
    with open("./data/candidatefiles100/" + ds_name + ".json", "r", encoding="utf-8") as json_file:
        data = json.load(json_file)
        # documentcontextmap = preparedocumentcontextmap(entity_embeddings, data)
        for i in data:
            sentences.append([entry.is_string for entry in document_dictionary.get(i).nifContent if
                              entry.uri == i][0])
        sentence_embeddings = prepare_sentence_embeddings(sentences)
        for i in data:
            for entity_entry in data[i]:
                candidate_vector = []
                gold_candidate = [entry.taIdentRef for entry in document_dictionary.get(i).nifContent if
                                  entry.uri == entity_entry['mention_uri']]
                if gold_candidate[0] in entity_embeddings:
                    y_train.append(np.array(entity_embeddings[gold_candidate[0]].split(","), dtype='float'))
                else:
                    y_train.append(np.full(100, -1))
                for candidate in entity_entry['candidates']:
                    if candidate in entity_embeddings:
                        candidate_vector.append(np.array(entity_embeddings[candidate].split(","), dtype='float'))
                    else:
                        candidate_vector.append(np.full(100, -1))
                if len(candidate_vector) < 100:
                    while len(candidate_vector) < 100:
                        candidate_vector.append(np.full(100, -1))
                else:
                    candidate_vector = candidate_vector[:100]
                x_train.append(candidate_vector)
                doc_embeddings.append(
                    sentence_embeddings[[entry.is_string for entry in document_dictionary.get(i).nifContent if
                                         entry.uri == i][0]])
                # comtextmatrices.append(documentcontextmap[i])
        return x_train, y_train, doc_embeddings, comtextmatrices


def get_ref_entity(candidates, prediction):
    sim = tf.keras.losses.cosine_similarity(np.array(candidates, dtype='double'), np.array(prediction, dtype='double'),
                                            axis=-1)
    return candidates[np.argmin(sim.numpy())]


def get_entity_from_embeddings(embeddings_array, embedding_df):
    entities_list = []
    for embedding in embeddings_array:
        entity = embedding_df[embedding_df.embeddings == ",".join(map(str, embedding))]
        if entity.empty:
            entities_list.append("UNKNOWN")
        else:
            entities_list.append(entity.get('entity').array[0])
    return entities_list


def evaluate(x_test, y_test, doc_emb, context):
    #doc_emb = np.reshape(doc_emb, (len(doc_emb), 1, 768))
    y_prediction = entity_linker([x_test, doc_emb])
    y_pred = []
    i = 0
    for entity in x_test:
        ref_entity = get_ref_entity(entity, y_prediction[i])
        y_pred.append(ref_entity)
        i += 1
    test_labels = get_entity_from_embeddings(y_test, data)
    predicted_labels = get_entity_from_embeddings(y_pred, data)
    in_kb_test_entities = []
    in_kb_pred_entities = []
    with open("./data/testb_compare.tsv", "w") as fl:
        for i in range(len(test_labels)):
            if str(test_labels[i]).find("dbpedia.org") >= 0 or str(predicted_labels[i]).find("dbpedia.org") >= 0:
                in_kb_test_entities.append(test_labels[i])
                in_kb_pred_entities.append(predicted_labels[i])
    print("Micro F1:\t" + str(f1_score(y_true=test_labels, y_pred=predicted_labels, average="micro")))
    print("In KB Micro F1:\t" + str(f1_score(y_true=in_kb_test_entities, y_pred=in_kb_pred_entities, average="micro")))


def rmsle(y_true, y_pred):
    msle = MeanSquaredError()
    return K.sqrt(msle(y_true, y_pred))


def train_model(train_x, train_y, train_doc_embedding, context):
    entity_in = Input(shape=(100, 100))
    doc_embedding_in = Input(shape=768)
    doc_embedding_dense = Dense(100, activation="relu")(doc_embedding_in)
    candidates = Flatten()(entity_in)
    candidate_embedding = Dense(100, activation="relu")(candidates)
    # candidate_embedding_conv = Flatten()(Conv1D(1, 1, activation="relu")(entity_in))
    combined_layer = concatenate([candidate_embedding, doc_embedding_dense])

    relu_layer = ReLU()(combined_layer)
    dropout_layer = Dropout(rate=0.2)(relu_layer)
    out_layer = Dense(100, activation="relu")(dropout_layer)
    model = Model([entity_in, doc_embedding_in], out_layer)
    optimizer = Adam(learning_rate=0.015)
    model.compile(loss=CosineSimilarity(axis=1), optimizer=optimizer, metrics=['accuracy'])
    num_epochs = 100
    # train_doc_embedding = np.reshape(train_doc_embedding, (len(train_doc_embedding), 1, 768))
    model.fit([train_x, train_doc_embedding], train_y, batch_size=128, epochs=num_epochs,
              validation_split=0.2, verbose=1)
    return model


data, embeddings = prepare_embeddings()
if exists("./models/linker_aida_train"):
    entity_linker = tensorflow.keras.models.load_model("./models/linker_aida_train")
    for ds_name in dataset_names:
        start = time.perf_counter()
        X_test, Y_test, doc_embedding_test, context = prepare_training_data(embeddings, ds_name)
        X_test = np.array(X_test)
        Y_test = np.array(Y_test)
        doc_embedding_test = np.array(doc_embedding_test)
        context = np.array(context)
        print("**************************")
        print(ds_name, "\n")
        evaluate(X_test, Y_test, doc_embedding_test, context)
        print("Time taken: ", time.perf_counter() - start)
        print("**************************")
else:
    x_train, y_train, doc_embedding, context = prepare_training_data(embeddings, 'aida_train')
    x_train = np.array(x_train)
    y_train = np.array(y_train)
    doc_embedding = np.array(doc_embedding)
    context = np.array(context)
    entity_linker = train_model(x_train, y_train, doc_embedding, context)
    entity_linker.save("./models/linker_aida_train")
    x_train = None
    y_train = None
    doc_embedding = None
